/*
 * Copyright (C) 2006,2007,2009  Kipp C. Cannon
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


/*
 * ============================================================================
 *
 *                           tokenizer Module API
 *
 * ============================================================================
 */


#include <Python.h>


#define MODULE_NAME "glue.ligolw.tokenizer"


/*
 * Classes
 */


extern PyTypeObject ligolw_Tokenizer_Type;
extern PyTypeObject ligolw_RowBuilder_Type;
extern PyTypeObject ligolw_RowDumper_Type;


/*
 * Functions
 */

PyObject *llwtokenizer_build_attributes(PyObject *sequence);
PyObject *llwtokenizer_build_formats(PyObject *sequence);
